void drawSentence() {
  for (int i = 0; i < bookRows.length; i++) {
    if (!bookRows[i].filtered(types[typeP], letter)) {
      color c = bookRows[i].getColor(types[typeP]);
      vec3f p = getPoint(bookRows[i].loc, rangeBook);
      stroke(c, 50);
      strokeWeight(5);
      pushMatrix();
        translate(p.x,p.y,p.z);
        point(0,0,0);
      popMatrix();
      strokeWeight(1);
      line(p.x, p.y, p.z, 0, 0, 0);
    }
  }
  hint(DISABLE_DEPTH_TEST);
  highlight h = new highlight(0,new vec3f(0,0,0));
  h = writeText(h, getPoint(bookRows[0].loc, rangeBook), bookRows[0].title, select, getPoint(bookRows[nearIndex].loc, rangeBook), 0);
  for (int i = 1; i< bookNumCols; i++) {
    if (!bookRows[i].filtered(types[typeP], letter)) {
      h = writeText(h, getPoint(bookRows[i].loc, rangeBook), bookRows[i].title, select, getPoint(bookRows[nearIndex].loc, rangeBook), i);
      if (near) { drawTitle(nearIndex, true); }
    }
  }
  if (h.count < select) { select = 0; }
  hint(ENABLE_DEPTH_TEST);
}

void drawLines(boolean draw) {
  String t = "";
  for (int i = 0; i < titleLookup.size(); i++) {
    if (!bookRows[i].filtered(types[typeP], letter)) {
      int b = drawTitle(i, draw);
      if (b > 50) {
        t = t + "\n" + bookRows[i].title.toUpperCase();
      }
    }
  }
  myTextlabelA.setText(t);
}


int drawTitle(int i, boolean draw) {
  ArrayList<Integer> innerList = titleLookup.get(i);
  color c = bookRows[i].getColor(types[typeP]);
  int b = contains(nearIndex, innerList);
  if (draw || b > 50) {
    switch(innerList.size()) {
      case 1:
        pushMatrix();
        vec3f pc = getPoint(wordRows[innerList.get(0)].loc, rangeWord);
        translate(pc.x, pc.y, pc.z);
        noFill();
        strokeWeight(5);
        if (near) {
          stroke(c, b);
        } else { stroke(c, 10); }
        point(0,0,0);
        popMatrix(); 
        break;
      default:
        for (int j = 0; j < innerList.size()-1; j++) {
          vec3f pc0 = getPoint(wordRows[innerList.get(j)].loc, rangeWord);
          vec3f pc1 = getPoint(wordRows[innerList.get(j+1)].loc, rangeWord);
          strokeWeight(2);
          if (near) {
            stroke(c, b);
          } else { stroke(c, 10); }
          line(pc0.x, pc0.y, pc0.z, 
               pc1.x, pc1.y, pc1.z);
        }
        break;
    }
  }
  return b;
}

void drawGeometry() {
  for (int i = 0; i < titleLookup.size(); i++) {
    if (!bookRows[i].filtered(types[typeP], letter)) {
      ArrayList<Integer> innerList = titleLookup.get(i);
      color c = bookRows[i].getColor(types[typeP]);
      switch(innerList.size()) {
        case 1:
          pushMatrix();
          vec3f pc = getPoint(wordRows[innerList.get(0)].loc, rangeWord);
          translate(pc.x, pc.y, pc.z);
          noFill();
          strokeWeight(5);
          stroke(c, 10); 
          point(0,0,0);
          popMatrix(); 
          break;
        case 2:
          vec3f pc0 = getPoint(wordRows[innerList.get(0)].loc, rangeWord);
          vec3f pc1 = getPoint(wordRows[innerList.get(1)].loc, rangeWord);
          strokeWeight(2);
          stroke(c, 10); 
          line(pc0.x, pc0.y, pc0.z, 
               pc1.x, pc1.y, pc1.z);
          break;
        case 3:
          vec3f pc0a = getPoint(wordRows[innerList.get(0)].loc, rangeWord);
          vec3f pc1a = getPoint(wordRows[innerList.get(1)].loc, rangeWord);
          vec3f pc2a = getPoint(wordRows[innerList.get(2)].loc, rangeWord);
          fill(c, 10);
          stroke(c, 10);
          beginShape();
          vertex(pc0a.x, pc0a.y, pc0a.z);
          vertex(pc1a.x, pc1a.y, pc1a.z);
          vertex(pc2a.x, pc2a.y, pc2a.z);
          endShape(CLOSE);
        case 0:
          break;
        default: //draw a 4 sided triangle
          vec3f pc0b = getPoint(wordRows[innerList.get(0)].loc, rangeWord);
          vec3f pc1b = getPoint(wordRows[innerList.get(1)].loc, rangeWord);
          vec3f pc2b = getPoint(wordRows[innerList.get(2)].loc, rangeWord);
          vec3f pc3b = getPoint(wordRows[innerList.get(3)].loc, rangeWord);
          fill(c, 10);
          stroke(c, 10); 
          beginShape(TRIANGLES);
          vertex(pc0b.x, pc0b.y, pc0b.z);
          vertex(pc1b.x, pc1b.y, pc1b.z);
          vertex(pc2b.x, pc2b.y, pc2b.z);
          
          vertex(pc0b.x, pc0b.y, pc0b.z);
          vertex(pc1b.x, pc1b.y, pc1b.z);
          vertex(pc3b.x, pc3b.y, pc3b.z);
          
          vertex(pc2b.x, pc2b.y, pc2b.z);
          vertex(pc1b.x, pc1b.y, pc1b.z);
          vertex(pc3b.x, pc3b.y, pc3b.z);
          
          vertex(pc2b.x, pc2b.y, pc2b.z);
          vertex(pc0b.x, pc0b.y, pc0b.z);
          vertex(pc3b.x, pc3b.y, pc3b.z);
          endShape(CLOSE);
      }
    }
  }
}

void drawWords() {
  for (int i = 0; i< wordNumCols; i++) {
    vec3f pc = getPoint(wordRows[i].loc, rangeWord);
    color c = color(0,0,0);
    strokeWeight(2);
    if (!geometry) {
      c = color(255,255,255);
      strokeWeight(5);
    }
    stroke(c,50);
    pushMatrix();
    translate(pc.x,pc.y,pc.z);
    point(0,0,0);
    popMatrix();
  }
}

void drawWordCloud() {
  hint(DISABLE_DEPTH_TEST);
  highlight h = new highlight(0,new vec3f(0,0,0));
  h = writeText(h, getPoint(wordRows[0].loc, rangeWord), wordRows[0].word, select, getPoint(wordRows[nearIndex].loc, rangeWord), 0);
  for (int i = 1; i< wordNumCols; i++) {
    h = writeText(h, getPoint(wordRows[i].loc, rangeWord), wordRows[i].word, select, getPoint(wordRows[nearIndex].loc, rangeWord), i);
  }
  if (h.count < select) { select = 0; }
  hint(ENABLE_DEPTH_TEST);
}

highlight writeText(highlight c, vec3f p, String title, int selected, vec3f nearPoint, int i) {
  if (distance(p, nearPoint))
  {
    if ( c.count == 0 ) { c.loc = p; }
    pushMatrix();
      translate(p.x,p.y,p.z);
      strokeWeight(5);
      stroke(255,255,255,255);
      point(0,0,0);
    popMatrix();
    pushMatrix();
      float[] rotations = cam.getRotations();
      translate(c.loc.x,c.loc.y,c.loc.z);
      rotateX(rotations[0]);
      rotateY(rotations[1]);
      rotateZ(rotations[2]);
      translate(0,-c.count*4,0);
      stroke(0,0,0,0);
      fill(255,255,255);
      textFont(font, 2);
      if ( c.count == selected && !near) {
          nearIndex = i;
          fill(210,210,210);
          textFont(fontBold, 2);
      } else if ( near && nearIndex == i ) {
        textFont(fontBold, 2);
      }
      text(title.toUpperCase(),0,0);
    popMatrix();
    c.count++;
  }
  return c;
}